from dataclasses import (
    dataclass,
    field,
)
from fa_purity.json.factory import (
    JsonObj,
)
from jsonschema import (
    Draft4Validator,
)


@dataclass(frozen=True)
class InnerJsonSchema:
    raw: JsonObj
    validator: Draft4Validator = field(compare=False)



from fa_purity import (
    Cmd,
    JsonObj,
)
from fa_purity.json.transform import (
    dumps,
)
from fa_singer_io.singer import (
    SingerMessage,
)
from fa_singer_io.singer.record.core import (
    SingerRecord,
)
from fa_singer_io.singer.record.encode import (
    encode_record,
)
from fa_singer_io.singer.schema.core import (
    SingerSchema,
)
from fa_singer_io.singer.schema.encode import (
    encode_schema,
)
from fa_singer_io.singer.state.encode import (
    encode_state,
)
from typing import (
    IO,
)


def encode(singer: SingerMessage) -> JsonObj:
    if isinstance(singer, SingerRecord):
        return encode_record(singer)
    if isinstance(singer, SingerSchema):
        return encode_schema(singer)
    return encode_state(singer)


def emit(target: IO[str], singer: SingerMessage) -> Cmd[None]:
    def _action() -> None:
        target.write(dumps(encode(singer)))
        target.write("\n")

    return Cmd.from_cmd(_action)



from fa_singer_io.singer.record import (
    SingerRecord,
)
from fa_singer_io.singer.schema import (
    SingerSchema,
)
from fa_singer_io.singer.state import (
    SingerState,
)
from typing import (
    Union,
)

SingerMessage = Union[SingerRecord, SingerSchema, SingerState]


__all__ = [
    "SingerRecord",
    "SingerSchema",
    "SingerState",
]



from .core import (
    SingerRecord,
)
from fa_purity import (
    JsonObj,
    Maybe,
    UnfoldedJVal,
)
from fa_purity.frozen import (
    freeze,
)
from fa_purity.json.factory import (
    from_unfolded_dict,
)
from typing import (
    Dict,
)


def encode_record(record: SingerRecord) -> JsonObj:
    time_str = Maybe.from_optional(record.time_extracted).map(
        lambda date: date.to_utc_str()
    )
    raw_json: Dict[str, UnfoldedJVal] = {
        "type": "RECORD",
        "stream": record.stream,
        "record": record.record,
    }
    if time_str.map(bool).value_or(False):
        raw_json["time_extracted"] = time_str.unwrap()

    return from_unfolded_dict(freeze(raw_json))



from .arch import (
    forbidden_allowlist,
    project_dag,
    raise_or_return,
)
from arch_lint.dag.check import (
    check_dag_map,
    dag_map_completeness,
)
from arch_lint.forbidden import (
    check_forbidden,
)
from arch_lint.graph import (
    FullPathModule,
    ImportGraph,
)
from arch_lint.private import (
    check_private,
)

root = FullPathModule.assert_module("fa_singer_io")


def test_dag_creation() -> None:
    project_dag()


def test_dag() -> None:
    graph = ImportGraph.from_modules(root, False)
    check_dag_map(project_dag(), graph)


def test_dag_completeness() -> None:
    graph = ImportGraph.from_modules(root, False)
    dag_map_completeness(project_dag(), graph, root)


def test_forbidden_creation() -> None:
    forbidden_allowlist()


def test_forbidden() -> None:
    graph = ImportGraph.from_modules(root, True)
    allowlist_map = forbidden_allowlist()
    check_forbidden(allowlist_map, graph)


def test_private() -> None:
    graph = ImportGraph.from_modules(root, False)
    check_private(graph, root)



from .core import (
    SingerSchema,
)
from fa_purity import (
    JsonObj,
    JsonValue,
    Maybe,
    UnfoldedJVal,
)
from fa_purity.frozen import (
    freeze,
)
from fa_purity.json.factory import (
    from_unfolded_dict,
)
from typing import (
    Dict,
)


def encode_schema(schema: SingerSchema) -> JsonObj:
    bookmark_properties = (
        Maybe.from_optional(schema.bookmark_properties)
        .map(lambda s: freeze([JsonValue(item) for item in s]))
        .to_result()
        .to_union()
    )
    raw_json: Dict[str, UnfoldedJVal] = {
        "type": "SCHEMA",
        "stream": schema.stream,
        "schema": schema.schema.encode(),
        "key_properties": freeze(
            [JsonValue(item) for item in schema.key_properties]
        ),
    }
    if bookmark_properties is not None:
        raw_json["bookmark_properties"] = bookmark_properties
    return from_unfolded_dict(freeze(raw_json))



from .core import (
    SingerRecord,
)
from fa_purity import (
    Cmd,
    JsonObj,
)
from fa_singer_io.time import (
    DateTime,
)


def new_record_auto_time(stream: str, record: JsonObj) -> Cmd[SingerRecord]:
    return DateTime.now().map(lambda d: SingerRecord(stream, record, d))



from __future__ import (
    annotations,
)

from .._utils import (
    all_keys_in,
)
from dataclasses import (
    dataclass,
)
from fa_purity import (
    FrozenDict,
    Result,
    ResultE,
)
from fa_purity.json.value.transform import (
    Unfolder,
)
from fa_purity.pure_iter.factory import (
    from_flist,
)
from fa_singer_io.json_schema import (
    JSchemaFactory,
    JsonSchema,
)
from fa_singer_io.singer.errors import (
    MissingKeys,
)
from typing import (
    FrozenSet,
    Optional,
    Tuple,
)


def _check_keys(
    properties: Optional[FrozenSet[str]],
    key_properties: FrozenSet[str],
    bookmark_properties: Optional[FrozenSet[str]],
) -> ResultE[None]:
    if properties is None:
        return (
            Result.failure(
                Exception(
                    "If not properties then not key_properties nor bookmark_properties at SingerSchema"
                )
            )
            if (bool(key_properties) or bool(bookmark_properties))
            else Result.success(None)
        )
    check_keys = all_keys_in(properties, key_properties).alt(
        lambda m: MissingKeys(m, "schema properties")
    )
    check_bookmarks = all_keys_in(properties, bookmark_properties).alt(
        lambda m: MissingKeys(m, "schema properties")
    )
    return check_keys.bind(lambda _: check_bookmarks).alt(Exception)


@dataclass(frozen=True)
class _Private:
    pass


@dataclass(frozen=True)
class Property:
    schema: JsonSchema
    key_property: bool
    bookmark_property: bool


@dataclass(frozen=True)
class SingerSchema:
    _private: _Private
    stream: str
    schema: JsonSchema
    key_properties: FrozenSet[str]
    bookmark_properties: Optional[FrozenSet[str]]

    @staticmethod
    def new(
        stream: str,
        schema: JsonSchema,
        key_properties: FrozenSet[str],
        bookmark_properties: Optional[FrozenSet[str]],
    ) -> ResultE[SingerSchema]:
        raw = schema.encode()
        props = raw.get("properties", None)
        properties = (
            frozenset(Unfolder(props).to_json().unwrap())
            if props is not None
            else None
        )
        return _check_keys(
            properties, key_properties, bookmark_properties
        ).map(
            lambda _: SingerSchema(
                _Private(), stream, schema, key_properties, bookmark_properties
            )
        )

    @staticmethod
    def obj_schema(
        stream: str, props: FrozenDict[str, Property]
    ) -> SingerSchema:
        schema = FrozenDict({k: v.schema for k, v in props.items()})
        key_properties = (
            from_flist(tuple(props.items()))
            .filter(lambda k: k[1].key_property)
            .map(lambda t: t[0])
            .transform(lambda i: frozenset(i))
        )
        bookmark_properties = (
            from_flist(tuple(props.items()))
            .filter(lambda k: k[1].bookmark_property)
            .map(lambda t: t[0])
            .transform(lambda i: frozenset(i))
        )
        return SingerSchema.new(
            stream,
            JSchemaFactory.obj_schema(schema),
            key_properties,
            bookmark_properties,
        ).unwrap()



from fa_purity import (
    ResultE,
)
from fa_singer_io.singer.record.core import (
    SingerRecord,
)
from fa_singer_io.singer.schema.core import (
    SingerSchema,
)


def validate_record(
    schema: SingerSchema, record: SingerRecord
) -> ResultE[None]:
    return schema.schema.validate(record.record).alt(Exception)



from .core import (
    SingerState,
)
from fa_purity.frozen import (
    freeze,
)
from fa_purity.json import (
    JsonObj,
    JsonValue,
)


def encode_state(state: SingerState) -> JsonObj:
    return freeze(
        {
            "type": JsonValue("STATE"),
            "value": JsonValue(state.value),
        }
    )



from ._inner import (
    InnerJsonSchema,
)
from dataclasses import (
    dataclass,
)
from fa_purity import (
    JsonObj,
    Result,
)
from fa_purity.json.transform import (
    to_raw,
)
from jsonschema.exceptions import (
    ValidationError,
)


@dataclass(frozen=True)
class JsonSchema:
    _inner: InnerJsonSchema

    def validate(self, record: JsonObj) -> Result[None, ValidationError]:
        try:
            self._inner.validator.validate(to_raw(record))  # type: ignore[misc]
            return Result.success(None)
        except ValidationError as error:  # type: ignore[misc]
            return Result.failure(error)

    def encode(self) -> JsonObj:
        return self._inner.raw



from fa_purity import (
    JsonValue,
)
from fa_purity.frozen import (
    freeze,
)
from fa_singer_io.singer.state.core import (
    SingerState,
)
from fa_singer_io.singer.state.decoder import (
    build_state,
)
from fa_singer_io.singer.state.encode import (
    encode_state,
)


def test_inverse() -> None:
    state = SingerState(
        freeze({"state": JsonValue("some state")}),
    )
    assert state == build_state(encode_state(state)).unwrap()



from .._utils import (
    all_keys_in,
)
from .core import (
    SingerState,
)
from fa_purity.json.factory import (
    JsonObj,
)
from fa_purity.json.value.transform import (
    Unfolder,
)
from fa_purity.result import (
    Result,
    ResultE,
    UnwrapError,
)
from fa_singer_io.singer.errors import (
    MissingKeys,
)


def build_state(raw_json: JsonObj) -> ResultE[SingerState]:
    required_keys = frozenset({"type", "value"})
    check = (
        all_keys_in(frozenset(raw_json), required_keys)
        .alt(lambda m: MissingKeys(m, "raw singer state"))
        .to_union()
    )
    if check is not None:
        return Result.failure(Exception(check), SingerState)
    try:
        parsed_type = Unfolder(raw_json["type"]).to_primitive(str).unwrap()
        if parsed_type == "STATE":
            record = SingerState(
                Unfolder(raw_json["value"]).to_json().unwrap(),
            )
            return Result.success(record)
    except UnwrapError as err:
        return Result.failure(err.container.unwrap_fail())
    return Result.failure(Exception(f'Expected "RECORD" not "{parsed_type}"'))



from fa_purity import (
    JsonValue,
)
from fa_purity.cmd import (
    unsafe_unwrap,
)
from fa_purity.frozen import (
    freeze,
)
from fa_singer_io.singer.record.decoder import (
    build_record,
)
from fa_singer_io.singer.record.encode import (
    encode_record,
)
from fa_singer_io.singer.record.factory import (
    new_record_auto_time,
)


def test_inverse() -> None:
    record = unsafe_unwrap(
        new_record_auto_time(
            "test_stream",
            freeze({"data": JsonValue(123)}),
        )
    )
    assert record == build_record(encode_record(record)).unwrap()



from json import (
    dump,
    load,
)
from pathlib import (
    Path,
)
import sys
from typing import (
    Tuple,
)


def set_settings(settings_path: Path, python_env: Path) -> None:
    settings_path.touch(exist_ok=True)
    env_settings = {
        "python.pythonPath": (python_env / "bin/python").as_posix(),
        "python.defaultInterpreterPath": (
            python_env / "bin/python"
        ).as_posix(),
        "mypy.dmypyExecutable": (python_env / "bin/dmypy").as_posix(),
        "python.linting.pylintPath": (python_env / "bin/pylint").as_posix(),
        "python.testing.pytestArgs": ["tests"],
        "python.testing.unittestEnabled": False,
        "python.testing.pytestEnabled": True,
        "python.linting.enabled": True,
        "python.linting.mypyEnabled": False,
        "python.linting.pylintEnabled": True,
    }
    with open(settings_path, "r+", encoding="UTF-8") as file:
        if file.read() == "":
            settings = {}
        else:
            file.seek(0)
            settings = load(file)  # type: ignore[misc]
    with open(settings_path, "w+", encoding="UTF-8") as file:
        new_settings = settings | env_settings
        dump(new_settings, file, indent=2)


def main(args: Tuple[str, ...]) -> None:
    settings = Path(args[1])
    env = Path(args[2])
    if env.exists():
        set_settings(settings, env)
    else:
        raise Exception(f"Non existent env: {env}")


if __name__ == "__main__":
    main(tuple(sys.argv))



from dataclasses import (
    dataclass,
)
from fa_purity import (
    Cmd,
    JsonObj,
    PureIter,
    Result,
    ResultE,
)
from fa_purity.json.factory import (
    loads,
)
from fa_purity.json.value.transform import (
    Unfolder,
)
from fa_purity.pure_iter.factory import (
    unsafe_from_cmd,
)
from fa_purity.pure_iter.transform import (
    filter_opt,
)
from fa_purity.union import (
    inl,
)
from fa_singer_io.singer import (
    SingerMessage,
)
from fa_singer_io.singer.errors import (
    MissingKeys,
)
from fa_singer_io.singer.record.decoder import (
    build_record,
)
from fa_singer_io.singer.schema.decoder import (
    build_schema,
)
from fa_singer_io.singer.state.decoder import (
    build_state,
)
from typing import (
    IO,
    Iterable,
)


@dataclass(frozen=True)
class _TmpWrapper:
    value: SingerMessage


def deserialize(raw: JsonObj) -> ResultE[SingerMessage]:
    parsed_type = (
        Unfolder(raw["type"])
        .to_primitive(str)
        .alt(lambda _: MissingKeys(frozenset(["type"]), "raw singer msg"))
    )
    _parsed_type = parsed_type.to_union()
    if _parsed_type == "RECORD":
        return build_record(raw).map(inl)
    if _parsed_type == "SCHEMA":
        return build_schema(raw).map(inl)
    if _parsed_type == "STATE":
        return build_state(raw).map(inl)
    return Result.failure(
        parsed_type.map(lambda t: Exception(f"Unknown type '{t}'"))
        .alt(Exception)
        .to_union()
    )


def _read_file(file: IO[str]) -> PureIter[str]:
    # IO file is supposed to be read-only
    def _iter_lines() -> Iterable[str]:
        line = file.readline()
        while line:
            yield line
            line = file.readline()

    return unsafe_from_cmd(Cmd.from_cmd(lambda: iter(_iter_lines())))


def try_from_file(file: IO[str]) -> PureIter[ResultE[SingerMessage]]:
    return (
        _read_file(file)
        .map(loads)
        .map(lambda r: r.alt(Exception).bind(deserialize))
    )


def from_file(file: IO[str]) -> PureIter[SingerMessage]:
    return try_from_file(file).map(lambda r: r.unwrap())


def from_file_ignore_failed(file: IO[str]) -> PureIter[SingerMessage]:
    return (
        try_from_file(file)
        .map(lambda r: r.map(_TmpWrapper).value_or(None))
        .transform(lambda i: filter_opt(i))
        .map(lambda w: w.value)
    )



from ._factory import (
    JSchemaFactory,
    SupportedType,
)
import warnings

warnings.warn(
    "`json_schema.factory` module is deprecated. Use `json_schema.JSchemaFactory` class instead",
    DeprecationWarning,
    stacklevel=2,
)

from_json = JSchemaFactory.from_json
multi_type = JSchemaFactory.multi_type
from_prim_type = JSchemaFactory.from_prim_type
opt_prim_type = JSchemaFactory.opt_prim_type
datetime_schema = JSchemaFactory.datetime_schema
opt_datetime_schema = JSchemaFactory.opt_datetime_schema

__all__ = [
    "SupportedType",
]



from arch_lint.dag import (
    DagMap,
)
from arch_lint.graph import (
    FullPathModule,
)
from fa_purity import (
    FrozenList,
)
from typing import (
    Dict,
    FrozenSet,
    TypeVar,
    Union,
)

_dag: Dict[str, FrozenList[Union[FrozenList[str], str]]] = {
    "fa_singer_io": (
        "singer",
        "json_schema",
        "time",
    ),
    "fa_singer_io.singer": (
        ("deserializer", "emitter", "validate", "encoder"),
        ("record", "schema", "state"),
        "errors",
        "_utils",
    ),
    "fa_singer_io.json_schema": (
        "factory",
        "_factory",
        "core",
        "_inner",
    ),
    "fa_singer_io.singer.state": (
        ("encode", "decoder"),
        "core",
    ),
    "fa_singer_io.singer.schema": (
        ("encode", "decoder", "factory"),
        "core",
    ),
    "fa_singer_io.singer.record": (
        ("encode", "decoder", "factory"),
        "core",
    ),
}
_T = TypeVar("_T")


def raise_or_return(item: Union[Exception, _T]) -> _T:
    if isinstance(item, Exception):
        raise item
    return item


def project_dag() -> DagMap:
    return raise_or_return(DagMap.new(_dag))


def forbidden_allowlist() -> Dict[FullPathModule, FrozenSet[FullPathModule]]:
    _raw: Dict[str, FrozenSet[str]] = {}
    return {
        raise_or_return(FullPathModule.from_raw(k)): frozenset(
            raise_or_return(FullPathModule.from_raw(i)) for i in v
        )
        for k, v in _raw.items()
    }



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
)
from fa_purity import (
    FrozenDict,
    JsonValue,
    PureIter,
)
from fa_purity.json.primitive import (
    Primitive,
)
from fa_singer_io.json_schema.core import (
    JsonSchema,
)
from fa_singer_io.singer import (
    SingerRecord,
    SingerSchema,
)
from fa_singer_io.singer.schema.core import (
    Property,
    SingerSchema,
)
from typing import (
    Callable,
    Generic,
    Optional,
    Type,
    TypeVar,
)

_T = TypeVar("_T")


@dataclass(frozen=True)
class _Patch(Generic[_T]):
    inner: _T


@dataclass(frozen=True)
class EncodeItem(Generic[_T]):
    _encode: _Patch[Callable[[_T], Primitive]]
    prop: Property

    def encode(self, item: _T) -> Primitive:
        return self._encode.inner(item)

    @staticmethod
    def new(
        encode: Callable[[_T], Primitive],
        prop: Property,
        _type: Optional[Type[_T]] = None,
    ) -> EncodeItem[_T]:
        """Use `_type` arg if typevar `_T` cannot be deduced and/or had to be set explicitly"""
        return EncodeItem(_Patch(encode), prop)


@dataclass(frozen=True)
class SingerEncoder(Generic[_T]):
    stream: str
    schema: SingerSchema
    _records: _Patch[Callable[[_T], SingerRecord]]

    def record(self, item: _T) -> SingerRecord:
        return self._records.inner(item)

    @staticmethod
    def new(
        stream: str, mapper: FrozenDict[str, EncodeItem[_T]]
    ) -> SingerEncoder[_T]:
        props = FrozenDict({k: v.prop for k, v in mapper.items()})
        schema = SingerSchema.obj_schema(stream, props)

        def _records(item: _T) -> SingerRecord:
            encoded = FrozenDict(
                {k: JsonValue(v.encode(item)) for k, v in mapper.items()}
            )
            return SingerRecord(stream, encoded, None)

        return SingerEncoder(
            stream,
            schema,
            _Patch(_records),
        )


@dataclass(frozen=True)
class FullSingerEncoder(Generic[_T]):
    obj_encoders: PureIter[SingerEncoder[_T]]

    @property
    def schemas(self) -> PureIter[SingerSchema]:
        return self.obj_encoders.map(lambda e: e.schema)

    def records(self, item: _T) -> PureIter[SingerRecord]:
        return self.obj_encoders.map(lambda e: e.record(item))



from dataclasses import (
    dataclass,
)
from fa_purity.json.factory import (
    JsonObj,
)
from fa_singer_io.time import (
    DateTime,
)
from typing import (
    Optional,
)


@dataclass(frozen=True)
class SingerRecord:
    stream: str
    record: JsonObj
    time_extracted: Optional[DateTime]



from .core import (
    SingerSchema,
)
from fa_singer_io.json_schema.core import (
    JsonSchema,
)


def from_jschema(
    stream: str,
    schema: JsonSchema,
) -> SingerSchema:
    return SingerSchema.new(stream, schema, frozenset([]), None).unwrap()



from fa_purity import (
    JsonValue,
)
from fa_purity.frozen import (
    freeze,
)
from fa_purity.json.factory import (
    from_unfolded_dict,
)
from fa_purity.json.primitive.core import (
    PrimitiveTypesList,
)
from fa_singer_io.json_schema.factory import (
    datetime_schema,
    from_json,
    from_prim_type,
    opt_datetime_schema,
    opt_prim_type,
)

MOCK_SCHEMA = from_json(
    from_unfolded_dict(
        freeze(
            {
                "properties": freeze(
                    {"foo": JsonValue(from_prim_type(int).encode())}
                )
            }
        )
    )
)


def test_mock_schema() -> None:
    valid = freeze({"foo": JsonValue(123)})
    invalid = freeze({"foo": JsonValue("text")})
    assert MOCK_SCHEMA.unwrap().validate(valid).unwrap() is None
    assert MOCK_SCHEMA.unwrap().validate(invalid).unwrap_fail()


def test_from_prim_type() -> None:
    for t in PrimitiveTypesList:
        assert from_prim_type(t)


def test_opt_prim_type() -> None:
    for t in PrimitiveTypesList:
        assert opt_prim_type(t)


def test_datetime_schema() -> None:
    assert datetime_schema()


def test_opt_datetime_schema() -> None:
    assert opt_datetime_schema()



from ._inner import (
    InnerJsonSchema,
)
from dataclasses import (
    dataclass,
)
from decimal import (
    Decimal,
)
from enum import (
    Enum,
)
from fa_purity import (
    FrozenDict,
    FrozenList,
    JsonObj,
    JsonValue,
    Result,
    ResultE,
)
from fa_purity.json.primitive import (
    PrimitiveTypes,
)
from fa_purity.json.transform import (
    to_raw,
)
from fa_singer_io.json_schema.core import (
    JsonSchema,
)
from jsonschema import (
    Draft4Validator,
    SchemaError,
)
from typing import (
    FrozenSet,
)


class SupportedType(Enum):
    array = "array"
    boolean = "boolean"
    integer = "integer"
    null = "null"
    number = "number"
    object = "object"
    string = "string"


_encode_type = {
    bool: SupportedType.boolean,
    int: SupportedType.integer,
    type(None): SupportedType.null,
    Decimal: SupportedType.number,
    float: SupportedType.number,
    str: SupportedType.string,
}


@dataclass(frozen=True)
class JSchemaFactory:
    @staticmethod
    def from_json(raw: JsonObj) -> ResultE[JsonSchema]:
        raw_dict = to_raw(raw)  # type: ignore[misc]
        try:
            Draft4Validator.check_schema(raw_dict)  # type: ignore[misc]
            validator = Draft4Validator(raw_dict)  # type: ignore[misc]
            draft = InnerJsonSchema(raw, validator)
            return Result.success(JsonSchema(draft))
        except SchemaError as err:  # type: ignore[misc]
            return Result.failure(err)

    @classmethod
    def multi_type(
        cls, types: FrozenSet[PrimitiveTypes]
    ) -> ResultE[JsonSchema]:
        if len(types) == 0:
            return Result.failure(Exception("Must specify a type"))
        _types: FrozenList[JsonValue] = tuple(
            JsonValue(_encode_type[t].value) for t in types
        )
        raw = {"type": JsonValue(_types) if len(_types) > 1 else _types[0]}
        return Result.success(cls.from_json(FrozenDict(raw)).unwrap())

    @classmethod
    def from_prim_type(cls, p_type: PrimitiveTypes) -> JsonSchema:
        raw = {"type": JsonValue(_encode_type[p_type].value)}
        return cls.from_json(FrozenDict(raw)).unwrap()

    @classmethod
    def opt_prim_type(cls, p_type: PrimitiveTypes) -> JsonSchema:
        return cls.multi_type(frozenset([p_type, type(None)])).unwrap()

    @classmethod
    def datetime_schema(cls) -> JsonSchema:
        json = {
            "type": JsonValue(_encode_type[str].value),
            "format": JsonValue("date-time"),
        }
        return cls.from_json(FrozenDict(json)).unwrap()

    @classmethod
    def opt_datetime_schema(cls) -> JsonSchema:
        base = cls.opt_prim_type(str).encode()
        json = {
            "type": base["type"],
            "format": JsonValue("date-time"),
        }
        return cls.from_json(FrozenDict(json)).unwrap()

    @staticmethod
    def obj_schema(props: FrozenDict[str, JsonSchema]) -> JsonSchema:
        _props = FrozenDict(
            {k: JsonValue(v.encode()) for k, v in props.items()}
        )
        raw = {
            "properties": JsonValue(_props),
            "required": JsonValue(tuple(JsonValue(k) for k in _props.keys())),
        }
        return JSchemaFactory.from_json(FrozenDict(raw)).unwrap()



from fa_singer_io.json_schema.factory import (
    datetime_schema,
)
from fa_singer_io.singer.schema.core import (
    SingerSchema,
)
from fa_singer_io.singer.schema.decoder import (
    build_schema,
)
from fa_singer_io.singer.schema.encode import (
    encode_schema,
)
from fa_singer_io.singer.schema.factory import (
    from_jschema,
)
from tests.test_json_schema import (
    MOCK_SCHEMA,
)


def test_new_schema() -> None:
    assert from_jschema("foo", datetime_schema())


def test_inverse() -> None:
    schema = SingerSchema.new(
        "test_stream",
        MOCK_SCHEMA.unwrap(),
        frozenset(["foo"]),
        frozenset(),
    ).unwrap()
    assert schema == build_schema(encode_schema(schema)).unwrap()



from fa_purity import (
    Result,
)
from typing import (
    FrozenSet,
    Optional,
    TypeVar,
)

_T = TypeVar("_T")


def all_keys_in(
    items: FrozenSet[_T], requires: Optional[FrozenSet[_T]]
) -> Result[None, FrozenSet[_T]]:
    _requires: FrozenSet[_T] = (
        requires if requires is not None else frozenset([])
    )
    if all([r in items for r in _requires]):
        return Result.success(None)
    return Result.failure(_requires - items)



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
    field,
)
from datetime import (
    datetime,
    timezone,
)
from fa_purity.cmd import (
    Cmd,
)
from fa_purity.date_time import (
    DatetimeFactory,
    DatetimeTZ,
    DatetimeUTC,
)
from pyrfc3339 import (
    generate,
    parse as parse_rfc3339,
)


@dataclass(frozen=True)
class _Private:
    pass


def _no_micro(date: datetime) -> datetime:
    return datetime(
        date.year,
        date.month,
        date.day,
        date.hour,
        date.minute,
        date.second,
        0,
        date.tzinfo,
    )


@dataclass(frozen=True)
class DateTime:
    _private: _Private = field(repr=False, hash=False, compare=False)
    _date: DatetimeUTC

    @staticmethod
    def from_utc(date: DatetimeUTC) -> DateTime:
        return DateTime(
            _Private(),
            DatetimeUTC.assert_utc(_no_micro(date.date_time)).unwrap(),
        )

    @classmethod
    def now(cls) -> Cmd[DateTime]:
        return DatetimeFactory.date_now().map(cls.from_utc)

    @classmethod
    def parse(cls, raw: str) -> DateTime:
        return cls.from_utc(
            DatetimeFactory.to_utc(
                DatetimeTZ.assert_tz(parse_rfc3339(raw)).unwrap()
            )
        )

    def to_utc_str(self) -> str:
        return generate(self._date.date_time)

    def to_str(self) -> str:
        return generate(self._date.date_time, utc=False)



from .._utils import (
    all_keys_in,
)
from .core import (
    SingerSchema,
)
from fa_purity import (
    JsonObj,
    Maybe,
    Result,
    ResultE,
)
from fa_purity.json.value.transform import (
    Unfolder,
)
from fa_purity.result import (
    UnwrapError,
)
from fa_singer_io.json_schema.factory import (
    from_json,
)
from fa_singer_io.singer.errors import (
    MissingKeys,
)


class DecodeError(Exception):
    pass


def build_schema(raw_json: JsonObj) -> ResultE[SingerSchema]:
    required_keys = frozenset({"type", "stream", "schema", "key_properties"})
    check = (
        all_keys_in(frozenset(raw_json), required_keys)
        .alt(lambda m: MissingKeys(m, "raw singer schema"))
        .to_union()
    )
    if check is not None:
        return Result.failure(check)
    parsed_type = Unfolder(raw_json["type"]).to_primitive(str).to_union()
    if parsed_type == "SCHEMA":
        try:
            bookmark_properties = (
                Maybe.from_optional(raw_json.get("bookmark_properties", None))
                .map(Unfolder)
                .map(lambda item: item.to_list_of(str).unwrap())
                .map(lambda item: frozenset(item))
            )
            return SingerSchema.new(
                Unfolder(raw_json["stream"]).to_primitive(str).unwrap(),
                from_json(
                    Unfolder(raw_json["schema"]).to_json().unwrap()
                ).unwrap(),
                frozenset(
                    Unfolder(raw_json["key_properties"])
                    .to_list_of(str)
                    .unwrap()
                ),
                bookmark_properties.value_or(None),
            )
        except UnwrapError as err:
            return Result.failure(err)
    return Result.failure(
        DecodeError(f'Expected "SCHEMA" not "{parsed_type}"')
    )
